#include <string>
#include <vector>
#include <curl/curl.h>
#include <thread>
#include <iostream>
#include <utility>
#include <algorithm>
#include "page.h"
#include "delimiter.h"
using namespace std;
size_t write_to_string(void *ptr, size_t size, size_t nmemb, string stream)
{
	size_t realsize = size * nmemb;
	string temp(static_cast<const char*>(ptr), realsize);
	stream.append(temp);
	return realsize;
}
page::page(const string& url, int description_max_length,
	int title_max_length, int paragraph_max_length):
	_description_max_length(description_max_length),
	_title_max_length(title_max_length),
	_paragraph_max_length(paragraph_max_length){
	CURL * curl;
	curl = curl_easy_init();

	if (curl) {
		string response;

		curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
		curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
		curl_easy_setopt(curl, CURLOPT_MAXREDIRS, 3L);
		curl_easy_setopt(curl, CURLOPT_TIMEOUT, 20L);
		curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 10L);
		curl_easy_setopt(curl, CURLOPT_USE_SSL, CURLUSESSL_TRY);
		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_to_string);

		curl_easy_setopt(curl, CURLOPT_WRITEDATA, &response);

		curl_easy_perform(curl);
		curl_easy_cleanup(curl);

		if(response.length() > 0){
			_source_page = istringstream(response);

		}else
			throw page::Invalid("Failed file retrieved " + url);
	}else
		curl_easy_cleanup(curl);
}

pair<string,string> page::next_link(){
	scrap_title();
	scrap_metainfo();
	pair<string, string> link_and_caption = scrap_url();
	if(link_and_caption.first.empty()){
		link_and_caption = scrap_robots_info();
		if(_tag_controller == 0)
			getline(_source_page, _line);
	}
	_tag_controller = 0;

	return link_and_caption;
}
void page::set_delimiter(const delimiter& delimiter){
	if(delimiter.get_type() == 0)
		_delimiters_link.push_back(delimiter);
	else if(delimiter.get_type() == 1)
		_delimiters_title.push_back(delimiter);
	else
		_delimiters_metainfo.push_back(delimiter);
}
void page::delete_delimiter(const delimiter& d_delimiter){
	if(d_delimiter.get_type() == 0){
		bool found = false;
	    vector<delimiter>::iterator it = _delimiters_link.begin();
		while(it != _delimiters_link.end() && !found){
			if((*it).get_start() != d_delimiter.get_start() &&
			    (*it).get_end() != d_delimiter.get_end()){
				++it;
			} else {
				found = true;
			}
		}
		_delimiters_link.erase(it);
	}
}
string page::get_metainfo(){
	string description = _description;
	string keywords = _keywords;
	if(description.empty()) description = _first_paragraph;
	if(keywords.size() > 0){
		size_t comma = 0;
		while(comma != string::npos){
			
			size_t next_comma = keywords.find(",",comma);
			string keyword;
			if(next_comma != string::npos){
				keyword = keywords.substr(comma, next_comma - comma);
			} else {
				keyword = keywords.substr(comma);
			}
			size_t has_keyword = description.find(keyword);
			keyword = delete_whitespaces(keyword);
			if(has_keyword == string::npos){
				description += " " + keyword;
			}
			if(next_comma != string::npos) ++next_comma;
			comma = next_comma;
		}
	}
	return description;
}
string page::delete_whitespaces(string line){
	line.erase( remove_if(begin(line), end(line), [l = locale{}](auto ch) {
		  return isspace(ch, l);
		}), end(line));

	return line;
}
void page::find_description(const string& url, const delimiter& delimit){
	size_t description_position = _line.find(delimit.get_start());
	if(description_position != string::npos){
		description_position += delimit.get_start().size();
		_line = _line.substr(description_position);
		size_t quote = _line.find(delimit.get_end());
		if(quote != string::npos){
			if(int(quote) > _description_max_length) quote = _description_max_length;
			_description = _line.substr(0, quote);
		}
	}
}

void page::scrap_title(){
	unsigned i = 0;
	while(_title.empty() && i < _delimiters_title.size()){
		size_t start_title_tag = _line.find(_delimiters_title[i].get_start());
		if(start_title_tag != string::npos){
			_line = _line.substr(start_title_tag + _delimiters_title[i].get_start().size());
			size_t end_title_tag = _line.find(_delimiters_title[i].get_end());
			if(end_title_tag != string::npos){
				string title = _line.substr(0, end_title_tag);
				if(int(title.size()) < _title_max_length){ _title = title;}
				else _title = title.substr(0, _title_max_length);
			}
			++_tag_controller;
		}
		++i;
	}
}
void page::scrap_metainfo(){
	unsigned i = 0;
	while(_description.empty() && i < _delimiters_metainfo.size()){
		find_description(_line, _delimiters_metainfo[i]);
		++i;
	}
	if(_keywords.empty()){
		size_t keywords_position = _line.find("<meta name=\"keywords\" content=\"");
		if(keywords_position != string::npos){
			_line = _line.substr(keywords_position + 31);
			size_t quote = _line.find("\"");
			if(quote != string::npos){
				if(int(quote) > _description_max_length) quote = _description_max_length;
				_keywords = _line.substr(0, quote);
			}
			++_tag_controller;
		}
	}
	if(_first_paragraph.empty() && _description.empty()){
		size_t p_tag = _line.find("<p>");
		if(p_tag != string::npos){
			string paragraph = _line.substr(p_tag + 3);
			size_t new_tag = paragraph.find('<');
			if(new_tag != string::npos){
				if(int(new_tag) < _paragraph_max_length)
					_first_paragraph = paragraph.substr(0, new_tag);
				else
					_first_paragraph = paragraph.substr(0, _paragraph_max_length);
			} else {
				if(int(paragraph.size()) < _paragraph_max_length)
					_first_paragraph = paragraph;
				else
					_first_paragraph = paragraph.substr(0, _paragraph_max_length);
			}
		}
	}
}
pair<string, string> page::scrap_url(){
	size_t href = _line.find("href=\"");
	
	bool has_href = false;
	pair<string,string> link_and_caption;
	//cout << _line << endl;
	if(href != string::npos){
		_line = _line.substr(href + 6);
		size_t first = _line.find("\"");
		if(first != string::npos){
			string line = _line.substr(0, first);
			string caption = "";
		 
			size_t hash = line.find("#");
			if(hash != string::npos)
				line = line.substr(0,hash);
			size_t semicolon = line.find(";");
			if(semicolon == string::npos){
				size_t curly_brace = line.find("{");
				if(curly_brace == string::npos){
					size_t angular = _line.find('>');
					if(angular != string::npos){
						++angular;
						string link_text = _line.substr(angular);
						size_t end = link_text.find("</a>");
						if(end != string::npos){
							caption = link_text.substr(0, end);
							angular = caption.find('>');
							if(angular != string::npos)
								caption.clear();
					
						}
					}
					has_href = true;
					pair<string, string> lac(line, caption);
					link_and_caption = lac;
				}
			}
		}
		++_tag_controller;
	}
	unsigned i = 0;
	while(i < _delimiters_link.size() && !has_href){
		size_t start_link_tag = _line.find(_delimiters_link[i].get_start());
		if(start_link_tag != string::npos){
			_line = _line.substr(start_link_tag + _delimiters_link[i].get_start().size());
			size_t end_link_tag = _line.find(_delimiters_link[i].get_end());
			if(end_link_tag != string::npos){
				string line = _line.substr(0, end_link_tag);

			 
				size_t hash = line.find("#");
				if(hash != string::npos)
					line = line.substr(0,hash);
				size_t semicolon = line.find(";");
				if(semicolon == string::npos){
					size_t curly_brace = line.find("{");
					if(curly_brace == string::npos){
						has_href = true;
						link_and_caption = pair<string, string>(line, "");
					}
				}
			}
			++_tag_controller;
		}
		++i;
	}
	return link_and_caption;
}
pair<string, string> page::scrap_robots_info(){
	size_t loc = _line.find("<loc>");
	pair<string,string> link_and_caption;
	if(loc != string::npos){
		_line = _line.substr(loc + 5);
		size_t angular = _line.find("<");
		_line = _line.substr(0, angular);
		pair<string, string> lac(_line, "");
		link_and_caption = lac;
		++_tag_controller;
	}else{
		size_t ref = _line.find("Sitemap: ");
		if(ref != string::npos){
			_line = _line.substr(ref + 9);
			pair<string, string> lac(_line,"");
			link_and_caption = lac;
			++_tag_controller;
		}
	}
	return link_and_caption;
}
