/*
* This file is the main code of Texdi
* Copyright (C) <2017>  <alkeon> [alkeon@autistici.org]
   
* Texdi is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Texdi is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Texdi.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include "tedi2ad.h"
#include "exception.h"
using namespace std;

#define BLOCK 2
#define IMAGE 1
#define LINK 0
#define NO_TAG -1
/*
 * This function converts a string with Tedi format and return 
 * a string with HTML format.
 * Header and footer are void by default.
 */
string tedi2ad(string text, string header, string footer){
	stringstream index(text);
	string line;
	string return_text = header;
	bool is_converting_table = false;
	int open_brackets=0;
	while(getline(index,line)){
		if(line[0]!='<'){
			size_t pipe = line.find("|");
			if(pipe == string::npos && is_converting_table){
				return_text += "|=======\n";
				is_converting_table = false;
			}
			size_t hash = line.find('#');
			if(hash!=std::string::npos){
				line=titles_ad(line);
				return_text += line+"\n";
			}else{
				size_t li = line.find("--");
				if(li!=string::npos){
					line=line.erase(li,2);
					
					line=line.insert(li,"- ");
				}
				size_t ul = line.find("__");
				size_t ulf = line.find(",,");
				if(ul != string::npos or ulf!=string::npos)
					return_text+="\n";
				size_t quotes = line.find("\"");
				if(quotes != string::npos){
					if(line[line.size()-1]=='"')
						line=line.substr(1,line.size()-2);
					else
						throw Invalid("Missing end quotes.",line);
				}else{
					int tag = main_tag_ad(line);
					while(tag != NO_TAG){
						if(tag == BLOCK){
					  		size_t lla2=line.find("}");
							if(lla2==std::string::npos){
								++open_brackets;
								line = incomplete_blocks_ad(line);
							}else
								line=blocks_ad(line);
						
						}else{
							if(tag == IMAGE){
								size_t par1 = line.find("([");
								if(par1 != std::string::npos )
									line=images_ad(line);
									
							}else
								line=links_ad(line);
								
						}
						tag = main_tag_ad(line);
					}
					size_t bracket2=line.find("}");
					if(bracket2 != string::npos){
						if(open_brackets==0)
							throw Invalid("Missing '{' in block tag.",line);
						--open_brackets;
						line = line.erase(bracket2,1);
					}
				}
					
				// Table conversion
				if(!is_converting_table){
					size_t last_pipe=line.rfind("|");
					if(pipe != string::npos && pipe != last_pipe){
						return_text += "|=======\n";
						table_ad(line);
						return_text += line + "\n";
						is_converting_table = true;
					}
				}else{
					if(pipe != string::npos){
						table_ad(line);
						return_text += line + "\n";
					}
				}
					
				if(ul==string::npos && ulf==string::npos && pipe == string::npos){
					size_t last_whitespace = line.rfind(" ");
					if(last_whitespace == line.size()-1 && line.size() > 1){
						return_text += line+ "\n\n";
					}else{
						return_text += line+ "\n";
					}
				}
			}		
		}else{
			if(line[1] != '!'){
				if(line[1] == '>'){
					line = line.erase(0,2);
					return_text += line + "\n\n";
				}else{
					line=line.erase(0, 1);
					line = line.insert(0, "`");
					return_text += line + "`\n\n";
				}
			}
		}
	}
	if(open_brackets > 0)
		throw Invalid("Missing '}' in document.","End of file");
	return_text += footer;

	return return_text;
}
/*
 * Detects next tag
 * Returns 
 * 0: images
 * 1: links
 * 2: blocks
 * -1: No tag
 */
int main_tag_ad(string line){
	int tag = NO_TAG;
	size_t size = line.size();
	size_t bracket = line.find("{(");
	size_t parenthesis = line.find("([");
	size_t square_bracket = line.find("[(");
	if(bracket != string::npos && bracket < size){
		tag = BLOCK;
		size = bracket;
	}
	if(square_bracket != string::npos && square_bracket < size){
		size = square_bracket;
		tag = LINK;
	}
	if(parenthesis != string::npos && parenthesis < size)
		tag = IMAGE;
	
	return tag;
}
/*
 * Deletes parentheses and adds ascii text of a image.
 *
 */
string images_ad(string line){

	size_t start_tag=line.find("([");
	if(start_tag != string::npos){
		line = line.erase(start_tag,2);
		size_t end_bracket = line.find("]");
		if(end_bracket != string::npos){
			if(line[end_bracket + 1] != ' ')
				throw Invalid("Missing ' ' in images tag.",line);
			string caption = line.substr(start_tag, end_bracket - start_tag);
			line = line.erase(start_tag, end_bracket + 2 - start_tag);
			line = line.insert(start_tag, "image::");

			size_t parenthesis = line.rfind(")");
			if(parenthesis != string::npos){
				line = line.erase(parenthesis, 1);
				line = line.insert(parenthesis, "[" + caption + "]");
			}else
				throw Invalid("Missing ')' in image tag.",line);
		}else
			throw Invalid("Missing ']' in image tag.",line);
	}else
		throw Invalid("Missing image tag.",line);
	return line;
}
/*
 * Convert Tedi link to a HTML link
 *
 */
string links_ad(string line){
	size_t start_tag = line.find("[(");
	if(start_tag != string::npos){
		line = line.erase(start_tag, 2);
		size_t par1 = line.find(")");
		if(par1 != string::npos){
			line = line.erase(par1,1);
			line = line.insert(par1, "[");
			size_t brack = line.rfind("]");
			if(brack == string::npos)
				throw Invalid("Missing ']' in link tag.",line);
		}else
			throw Invalid("Missing ')' in link tag.",line);
	}else
		throw Invalid("Missing link tag.",line);

	return line;
}
/*
 * Deletes hashes and adds HTML text of a title.
 *
 */
string titles_ad(string line){
	size_t exa1=line.find("#");
	int level=0;
	while(exa1!=string::npos){
		++exa1;
		line = line.substr(exa1,line.size()-1);
		exa1 = line.find("#");
		++level;
	}
	string aux=line;
	line+="\n";
	char let;
	if(level == 1)
		let = '-';
	else if(level == 2)
		let = '~';
	else if(level == 3)
		let = '^';
	else
		let = '+';
	for(unsigned i = 0 ;i < aux.length(); ++i){
		line += let;
	}
	line += "\n";
	return line;
}
/*
 * Deletes brackes and parentheses. Adds HTML text 
 * of a <div> container.
 *
 */
string blocks_ad(string line){
	size_t bracket=line.find("{(");

	if(bracket != std::string::npos){
		size_t parenthesis = line.find(")");
		if(parenthesis != string::npos){
			int delete_class = parenthesis - bracket;
			line = line.erase(bracket, delete_class);
			size_t end_bracket = line.rfind("}");
			if(end_bracket != string::npos)
				line = line.erase(end_bracket, 1);
			else
				throw Invalid("Missing '}' in block tag.",line);
		}else
			throw Invalid("Missing ')' in block tag.",line);
	}else
		throw Invalid("Missing block tag.",line);
	return line;
}
/*
 * Deletes brackes and parentheses. Adds HTML text 
 * of a <div> container.
 *
 */
string incomplete_blocks_ad(std::string line){
	size_t bracket = line.find("{(");
	if(bracket != std::string::npos){
		size_t parenthesis = line.find(")");
		if(parenthesis != string::npos){
			int delete_class = parenthesis - bracket;
			line = line.erase(bracket, delete_class);
			size_t end_bracket = line.find("}");
			if(end_bracket != string::npos)
				throw Invalid("Unexpected '}' in incomplete block tag.",line);
		}else
			throw Invalid("Missing ')' in block tag.",line);
	}else
		throw Invalid("Missing block tag.",line);
	return line;
}

/*
 * Convert Tedi table (line by line)
 *
 */
void table_ad(string& line){
	size_t pipe = line.find("|");
	if(pipe != string::npos){
		pipe = line.find("|", pipe);
		if(pipe != string::npos)
			line=line.erase(line.size()-1,1);
		else
			throw Invalid("Expected '|' in table", line);
		
	}else
		throw Invalid("Expected '|' in table", line);
	
}
