/*
* This file is the main code of Texdi
* Copyright (C) <2017>  <alkeon> [alkeon@autistici.org]
*
* Texdi is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Texdi is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Texdi.  If not, see <http://www.gnu.org/licenses/>.
*
*/
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include "tedi2html.h"
#include "exception.h"
using namespace std;

#define BLOCK 2
#define IMAGE 1
#define LINK 0
#define NO_TAG -1
/*
 * This function converts a string with Tedi format and return 
 * a string with HTML format.
 * Header and footer are by default void.
 */
string close_paragraph(bool& is_paragraph){
	string p("");
	if(is_paragraph){
		p = "</p>\n";
		is_paragraph = false;
	}
	return p;
}
string tedi2html(string text, string header, string footer){
	stringstream index(text);
	string line, return_text = header;
	bool is_converting_table = false;
	int open_brackets = 0;
	bool is_paragraph = false;
	bool unordered_list = false;
	while(getline(index, line)){
		bool posible_paragraph = true;
		if(line[0]!='<'){
			size_t pipe = line.find("|");
			if(pipe == string::npos && is_converting_table){
				return_text += "</table>\n";
				is_converting_table = false;
				posible_paragraph = false;
			}
			size_t hash =line.find('#');
			if(hash != string::npos){
				posible_paragraph = false;
				line = titles_ht(line);
				return_text += close_paragraph(is_paragraph) + line + "\n";
			}else{
				size_t ul = line.find("__");
				if(ul != string::npos){
					posible_paragraph = false;
					return_text += close_paragraph(is_paragraph) + "<ul>\n";
					unordered_list = true;
				}
				size_t li=line.find("--");
				if(li != string::npos){
					posible_paragraph = false;
					line = line.erase(li,2);
					line = line.insert(li,close_paragraph(is_paragraph) + "<li>");
				}
				size_t ul_end = line.find(",,");
				if(ul_end != string::npos){
					posible_paragraph = false;
					return_text += close_paragraph(is_paragraph) + "</ul>\n";
					unordered_list = false;
				}
				size_t quotes=line.find("\"");
				if(quotes != string::npos){

					if(line[line.size()-1] == '"'){
						line = line.substr(1,line.size()-2);
					}else
						throw Invalid("Missing end quotes.",line);
					
				}else{
					int tag = main_tag(line);
					while(tag != NO_TAG){
						if(tag == BLOCK){
							size_t end_bracket = line.find("}");
							if(end_bracket == string::npos){
								++open_brackets;
								line = incomplete_blocks_ht(line);
							}else
								line = blocks_ht(line);
							
						}else{
							if(tag == IMAGE)
								line = images_ht(line);
							else
								line = links_ht(line);
						}
						tag = main_tag(line);
					}
					size_t first_bracket_closed = line.find("}");
					if(first_bracket_closed != string::npos){
						if(open_brackets == 0)
							throw Invalid("Missing '{' in block tag.", line);
						--open_brackets;
						line = line.erase(first_bracket_closed, 1);
						line = line.insert(first_bracket_closed, "</div>");
					}
				}
					
				size_t first_pipe = line.find("|");
				if(!is_converting_table){
					if(first_pipe != string::npos && first_pipe != line.rfind("|")){
						posible_paragraph = false;
						return_text += close_paragraph(is_paragraph) + "<table>\n";
						table_ht(line, return_text);
						is_converting_table = true;
					}
				}else{
					if(first_pipe != string::npos){
						posible_paragraph = false;
						table_ht(line, return_text);
					}
				}
				if(ul == string::npos && ul_end == string::npos && first_pipe==string::npos){

					if(line.rfind(" ") == line.size()-1 && line.size() > 1) {
						if(is_paragraph) {
							return_text += line + "</p>\n";
							is_paragraph = false;
						} else
							return_text += "<p>" + line + "</p>\n";
					} else {
						if(line.size() > 1 && !unordered_list) {
							if(posible_paragraph) {
								if(is_paragraph)
									return_text += line + "\n";
								else {
									return_text += "<p>" + line + "\n";
									is_paragraph = true;
								}
							} else {
								if(is_paragraph) {
									return_text += "</p>\n" + line + "\n";
									is_paragraph = false;
								} else
									return_text += line + "\n";
							}
						} else
							return_text += line + "\n";
						
					}
				}
			}
		}else{
			if(line[1] != '!'){
				if(line[1] == '>'){
					line = line.erase(0,2);
					return_text += line + "\n";
				}else{
					line = line.erase(0,1);
					if(is_paragraph) {
						return_text += line + "</p>\n";
						is_paragraph = false;
					} else {
						return_text += "<p>" + line + "</p>\n";
					}
				}
			}
		}
	}
	if(open_brackets > 0)
		throw Invalid("Missing '}' in document.", "End of file");
	if(is_paragraph) return_text += "</p>\n";
	return_text += footer;

	return return_text;
}
/*
 * Detects next tag
 * Returns 
 * 0: images
 * 1: links
 * 2: blocks
 * -1: No tag
 */
int main_tag(string line){
	int tag = NO_TAG;
	size_t line_size = line.size();
	size_t first_bracket = line.find("{");
	size_t first_parenthesis = line.find("([");
	size_t first_square_bracket = line.find("[(");
	if(first_bracket != string::npos && first_bracket < line_size){
		tag = BLOCK;
		line_size = first_bracket;
	}
	if(first_square_bracket != string::npos && first_square_bracket < line_size){
		line_size = first_square_bracket;
		tag = LINK;
	}
	if(first_parenthesis != string::npos && first_parenthesis < line_size)
		tag = IMAGE;

	return tag;
}
/*
 * Convert Tedi table to a HTML link (line by line)
 *
 */
void table_ht(string& line, string& return_text){
	return_text += "<tr>\n";
	size_t pipe = line.find("|");
	if(pipe != string::npos){
		++pipe;
		line = line.substr(pipe, line.size()-1);
		pipe = line.find("|");
		if(pipe != string::npos){
			int size = line.size()-1;
			if(line[size] == '|'){
				while(pipe != string::npos && line[size]=='|'){
					return_text += "<td>" + line.substr(0, pipe) + "</td>\n";
					++pipe;
					line = line.substr(pipe, size);
					pipe = line.find("|");
					size = line.size()-1;
				}
				return_text += "</tr>\n";
			}else
				throw Invalid("Table not correctly written\n"
					"Maybe there's a whitespace at end of line", line);
		}else
			throw Invalid("Expected '|' in table", line);
	}else
		throw Invalid("Expected '|' in table", line);
}
/*
 * Convert Tedi link to a HTML link
 *
 */
string links_ht(string line){

	size_t start_tag = line.find("[(");
	if(start_tag != string::npos){
		line = line.erase(start_tag, 2);
		line = line.insert(start_tag, "<a href=\"");
		size_t end_parenthesis = line.find(")");
		if(end_parenthesis != string::npos){
			line = line.erase(end_parenthesis, 1);
			line = line.insert(end_parenthesis, "\">");
			size_t bracket = line.rfind("]");
			if(bracket != string::npos){
				line = line.erase(bracket, 1);
				line = line.insert(bracket, "</a>");
			}else
				throw Invalid("Missing ']' in link tag.", line);
		}else
			throw Invalid("Missing ')' in link tag.", line);
	}else
		throw Invalid("Missing link tag.", line);

	return line;
}
/*
 * Deletes image tag and adds HTML text of a image.
 *
 */
string images_ht(string line){
	size_t start_tag=line.find("([");
	if(start_tag!= string::npos){

		line = line.erase(start_tag, 2);
		line = line.insert(start_tag, "<img alt=\"");
		size_t square_bracket = line.find("]");

		if(square_bracket != string::npos){
			if(line[square_bracket + 1] != ' ')
				 throw Invalid("Missing ' ' in images tag.",line);
			line = line.erase(square_bracket, 2);
			line = line.insert(square_bracket, "\" src=\"");
			size_t last_bracket = line.rfind(")");
			if(last_bracket != string::npos){
				line = line.erase(last_bracket, 1);
				line = line.insert(last_bracket, "\">");
			}else
				throw Invalid("Missing ')' in images tag.",line);
		}else
			throw Invalid("Missing ']' in images tag.",line);
	}else
		throw Invalid("Missing images tag.",line);

	return line;
}
/*
 * Deletes hashes and adds HTML text of a title.
 *
 */
string titles_ht(string line){
	size_t exa1 = line.find("#");
	int level = 1;
	while(exa1 != string::npos){
		++exa1;
		line = line.substr(exa1,line.size()-1);
		exa1 = line.find("#");
		++level;
	}
	if(level > 6)
		level = 6;
	string level_s = to_string(level);
	line="<h" + level_s + ">" + line + "</h" + level_s + ">";
	return line;
}
/*
 * Deletes brackes and parentheses. Adds HTML text 
 * of a <div> container.
 *
 */
string blocks_ht(string line){
	size_t start_tag = line.find("{(");
	if(start_tag != string::npos){
		line = line.erase(start_tag, 2);
		line = line.insert(start_tag, "<div class=\"");
		start_tag = line.find(")");
		if(start_tag != string::npos){
			line = line.erase(start_tag, 1);
			line = line.insert(start_tag, "\">");
			start_tag = line.rfind("}");
			if(start_tag != string::npos){
				line = line.erase(start_tag, 1);
				line = line.insert(start_tag, "</div>");
			}else
				throw Invalid("Missing '}' in block tag.",line);
		}else
			throw Invalid("Missing ')' in block tag.",line);
	}else
		throw Invalid("Missing block tag.",line);
	return line;
}
/*
 * Deletes brackes and parentheses. Adds HTML text 
 * of a <div> container.
 *
 */
string incomplete_blocks_ht(string line){
	size_t start_tag = line.find("{(");
	if(start_tag != string::npos){

		line.erase(start_tag, 2);
		line.insert(start_tag, "<div class=\"");
		
		size_t parenthesis = line.find(")");
		if(parenthesis != string::npos){
			line.erase(parenthesis, 1);
			line.insert(parenthesis, "\">");
			size_t bracket = line.find("}");
			if(bracket != string::npos)
				throw Invalid("Unexpected '}' in incomplete block tag.",line);
		}else
			throw Invalid("Missing ')' in block tag.",line);
	}else
		throw Invalid("Missing block tag.",line);
	return line;
}
