/*
* This file is the main code of Texdi
* Copyright (C) <2017>  <alkeon> [alkeon@autistici.org]
   
* Texdi is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Texdi is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Texdi.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include "tedi2html.h"
#include "tedi2md.h"
#include "exception.h"
using namespace std;

#define BLOCK 2
#define IMAGE 1
#define LINK 0
#define NO_TAG -1
/*
 * This function converts a string with Tedi format and return 
 * a string with HTML format.
 * Header and footer are void by default.
 */
string tedi2md(string text, string header, string footer){
	stringstream index(text);
	string line, return_text = header;
	bool is_converting_table = false;
	int open_brackets = 0;
	while(getline(index,line)){
		if(line[0] != '<'){
			size_t pipe = line.find("|");
			if(pipe == string::npos && is_converting_table){
				return_text += "</table>\n";
				is_converting_table = false;
			}
			size_t hash = line.find('#');
			if(hash != std::string::npos){
		  		line = titles_md(line);
				return_text += line + "\n";
			}else{
				size_t ul = line.find("__");
				if(ul != string::npos)
					return_text += "<ul>\n";

				size_t li=line.find("--");
				if(li != string::npos){
					line = line.erase(li,2);
					line = line.insert(li,"<li>");
				}
				size_t ul_end = line.find(",,");
				if(ul_end != string::npos)
					return_text += "</ul>\n";
				
				size_t quotes = line.find("\"");
				if(quotes != string::npos){
					if(line[line.size()-1] == '"')
						line = line.substr(1, line.size()-2);
					else
						throw Invalid("Missing end quotes.",line);
				}else{
					int tag = main_tag(line);
					while(tag != NO_TAG){
						if(tag == BLOCK){
							size_t bracket = line.find("}");
							if(bracket == string::npos){
								++open_brackets;
								line = incomplete_blocks_ht(line);
							}else
							  line = blocks_ht(line);
							
						}else{
							if(tag == IMAGE){
								size_t parenthesis = line.find("(");
								if(parenthesis != std::string::npos)
							  		line = images_ht(line);
								
							}else
								line = links_ht(line);
							
						}
						tag = main_tag(line);
					}
					size_t end_bracket = line.find("}");
					if(end_bracket != string::npos){
						if(open_brackets == 0)
							throw Invalid("Missing '{' in block tag.",line);
						--open_brackets;
						line = line.erase(end_bracket, 1);
						line = line.insert(end_bracket, "</div>");
					}
				}
					
				// Table conversion
				size_t first_pipe = line.find("|");
				if(!is_converting_table){
					if(first_pipe != string::npos && first_pipe != line.rfind("|")){
						return_text += "<table>\n<tr>\n";
						table_ht(line, return_text);
						is_converting_table = true;
					}
				}else{
					if(first_pipe != string::npos){
						return_text += "<tr>\n";
						table_ht(line, return_text);
					}
				}
				if(ul_end==string::npos && ul==string::npos && first_pipe == string::npos){
					size_t last_whitespace = line.rfind(" ");
					if(last_whitespace == line.size()-1 && line.size()>1){
					  return_text += line + "<br>\n";
					}else{
					  return_text += line + "\n";
					}
				}
			}
		}else{
			if(line[1] != '!'){
				if(line[1] == '>'){
					line = line.erase(0,2);
					return_text += line + "<br>\n";
				}else{
					line = line.erase(0,1);
					return_text += line + "<br>\n";
				}
			}
		}
	}
	if(open_brackets > 0)
		throw Invalid("Missing '}' in document.","End of file");
	return_text += footer;
	return return_text;
}

/*
 * Deletes hashes and adds HTML text of a title.
 *
 */
string titles_md(string line){
	size_t exa1 = line.find("#");
	while(exa1 != string::npos){
		++exa1;
		line = line.substr(exa1,line.size()-1);
		exa1 = line.find("#");
	}
	line += "\n===\n";
	return line;
}
