/*
* This file is the main code of Texdi
* Copyright (C) <2016>  <alkeon> [alkeon@autistici.org]
   
* Texdi is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* Texdi is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Texdi.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <list>
#include "tedi2tex.h"
#include "exception.h"
using namespace std;

#define BLOCK 2
#define IMAGE 1
#define LINK 0
#define NO_TAG -1
/*
 * This function converts a string with Tedi format and return 
 * a string with Latex format.
 * Header and footer are void by default.
 */

string tedi2tex(string text, string header, string footer){
	list<int> list_table=info_tex(text);
	stringstream index(text);
	string line, return_text = header;
	bool is_converting_table = false;
	int open_brackets = 0;

	while(getline(index,line)){
		if(line[0] != '<'){
			//Change all _ to \_ Latex uses _ as underscore
			size_t underscore = line.find('_');
			while(underscore != string::npos){
				size_t ul2 = line.find("__");
				if(ul2 == string::npos)
					line = line.insert(underscore, 1, '\\');
				underscore = line.find('_', underscore + 2);
			}
			size_t pipe = line.find("|");
			if(pipe == string::npos && is_converting_table){
				return_text += "\\hline\n";
				return_text += "\\end{tabular}\n\n";
				is_converting_table = false;
			}
			size_t hash = line.find('#');
			if(hash != string::npos){
				line = title_tex(line);
				return_text += line + "\n";
			}else{
				size_t ul = line.find("__");
				if(ul != string::npos)
					return_text += "\\begin{itemize}\n";
				
				size_t li = line.find("--");
				if(li != string::npos){
					line = line.erase(li,2);
					line = line.insert(li,"\\item ");
				}
				size_t ulf = line.find(",,");
				if(ulf != string::npos)
					return_text += "\\end{itemize}\n";
				
				size_t quotes = line.find("\"");
				if(quotes != string::npos){
					if(line[line.size()-1] == '"')
						line = line.substr(1, line.size()-2);
					else
						throw Invalid("Missing end quotes.",line);
				}else{
					int tag = main_tag_tex(line);
					//size_t start_tag_div = line.find("{(");
					//if(bracket == string::npos && start_tag_div != string::npos)
					//	throw Invalid("Missing block tag.", line);
					//size_t bracket_tag = line.find("[(");
					//if(bracket_tag == string::npos && square_bracket != string::npos)
					//	throw Invalid("Missing link tag.",line);
					while(tag != NO_TAG){
						if(tag == BLOCK){
							size_t lla2 = line.find("}");
							if(lla2 != string::npos)
								line = blocks_tex(line);
							else{
								++open_brackets;
								line = incompletes_blocks_tex(line);
							}
						}else{
							if(tag == IMAGE){
						    		size_t par1=line.find("([");
								if(par1 != string::npos )
									line=images_tex(line);
								
							}else
								line=links_tex(line);
							
						}
						tag = main_tag_tex(line);
					}
					size_t close_bracket= line.find("}");
					size_t start_bracket=line.find("{");
					if(close_bracket != string::npos && start_bracket == string::npos){
						if(open_brackets==0)
							throw Invalid("Missing '{' in block tag.",line);
						--open_brackets;
						line = line.erase(close_bracket, 1);
					}
				}

				if(!is_converting_table){
					size_t last_pipe = line.rfind("|");
					if(pipe != string::npos){
						if(pipe != last_pipe){
							return_text += "\\begin{tabular}{";
							int i = 0;
							double size_per_column = 0.9 / list_table.front();
							while(i < list_table.front()){
								return_text += "|p{" + to_string(size_per_column) + "\\linewidth}";
								++i;
							}

							return_text += "|}\n";
							list_table.pop_front();
							table_tex(line, return_text);
							is_converting_table = true;
						}else
							throw Invalid("Missing '|' in table.",line);
						
					}
				}else{
					if(pipe != string::npos)
						table_tex(line, return_text);
						
				}


				if(ul == string::npos && ulf==string::npos && pipe==string::npos){
					if(line.rfind(" ") == line.size()-1 && line.size() > 1)
						return_text += line + "\n\n";
					else
						return_text += line + "\n";
					
				}
			}
		}else{
			if(line[1] != '!'){
				if(line[1] != '>'){
					line = line.erase(0, 1);
					return_text += "\\begin{verbatim}\n" + line + "\n\\end{verbatim}\n\n";
				}else{
					line = line.erase(0, 2);
					return_text += line + "\n\n";
				}
			}
		}
	}
	if(open_brackets > 0)
		throw Invalid("Missing '}' in document.","End of file");
	return_text += footer;

	return return_text;
}

/*
 * Detects next tag
 * Returns 
 * 0: images
 * 1: links
 * 2: blocks
 * -1: No tag
 */
int main_tag_tex(string line){
	int tag = -1;
	size_t size = line.size();
	size_t bracket = line.find("{(");
	size_t parenthesis = line.find("([");
	size_t square_bracket = line.find("[(");
	if(bracket != string::npos && bracket < size){
		tag = BLOCK;
		size = bracket;
	}
	if(square_bracket != string::npos && square_bracket < size){
		size = square_bracket;
		tag = LINK;
	}
	if(parenthesis != string::npos && parenthesis < size)
		tag = IMAGE;
	
	return tag;
}
/*
 * Convert Tedi table to a HTML link (line by line)
 *
 */
void table_tex(string& line, string& return_text){
	size_t pipe = line.find("|");
	if(pipe != string::npos){
		++pipe;
		return_text += "\\hline\n";
		line = line.substr(pipe, line.size()-1);
		pipe = line.find("|");
		if(pipe != string::npos){
			int size = line.size()-1;
			int x=0;
			if(line[size] == '|'){
				while(pipe != string::npos && line[size]=='|'){
					string table_data = line.substr(0, pipe);
					if(x != 0)
						return_text += " &" + table_data;
					else
						return_text += table_data;
			
					++pipe;
					line = line.substr(pipe, size);
					pipe = line.find("|");
					size = line.size()-1;
					++x;
				}
				return_text += "\\tabularnewline\n";
			}else
				throw Invalid("Table not correctly written\n"
					"Maybe there's a whitespace at end of line", line);
		}else
			throw Invalid("Expected '|' in table", line);
	}else
		throw Invalid("Expected '|' in table", line);
	
}
/*
 * Convert Tedi link to a Latex link
 *
 */
string links_tex(string line){
	size_t start_tag = line.find("[(");
	if(start_tag != string::npos){
		line=line.erase(start_tag, 2);
		line=line.insert(start_tag, "\\href{");
		size_t parenthesis=line.find(")");
		if(parenthesis != string::npos){
			line = line.erase(parenthesis, 1);
			line = line.insert(parenthesis, "}{");
			size_t brack = line.rfind("]");
			if(brack != string::npos){
				line = line.erase(brack, 1);
				line = line.insert(brack, "}");
			}else
				throw Invalid("Missing ']' in link tag.",line);
		}else
			throw Invalid("Missing ')' in link tag.",line);
	}else
		throw Invalid("Missing link tag.",line);

	return line;
}
/*
 * Deletes parentheses and adds Latex text of a image.
 *
 */
string images_tex(string line){
	size_t start_tag=line.find("([");
	if(start_tag != string::npos){
		line = line.erase(start_tag,2);
		size_t end_bracket = line.find("]");
		if(end_bracket != string::npos){
			if(line[end_bracket + 1] != ' ')
				throw Invalid("Missing ' ' in images tag.",line);
			string caption = line.substr(start_tag, end_bracket - start_tag);
			line = line.erase(start_tag, end_bracket + 2 - start_tag);
			line = line.insert(start_tag, "\\begin{figure}[h]\n\\centering\n"
				"\\includegraphics[width=0.5\\textwidth]{");

			size_t parenthesis = line.rfind(")");
			if(parenthesis != string::npos){
				line = line.erase(parenthesis, 1);
				line = line.insert(parenthesis, "}\n\\caption{" + caption + "}\n\\end{figure}");
			}else
				throw Invalid("Missing ')' in image tag.",line);
		}else
			throw Invalid("Missing ']' in image tag.",line);
	}else
		throw Invalid("Missing image tag.",line);
	return line;

}
/*
 * Deletes hashes and adds Latex text of a title.
 *
 */
string title_tex(string line){
	size_t exa1 =line.find('#');
	int level = -1;
	while(exa1 != string::npos){
		++exa1;
		line = line.substr(exa1, line.size()-1);
		exa1 = line.find('#');

		++level;
	}
	if(level == 0)
		line = "\\section{" + line + "}\n";
	else{
		if(level == 1)
			line = "\\subsection{" + line + "}\n";
		else
			line = "\\subsubsection{" + line + "}\n";
		
	}
	return line;
}
/*
 * Deletes brackes and parentheses.
 *
 */
string blocks_tex(string line){
	size_t start_tag=line.find("{(");

	if(start_tag != std::string::npos){
		line=line.erase(start_tag, 2);
		size_t parenthesis = line.find(")");
		if(parenthesis != string::npos){
			int delete_class = parenthesis - start_tag;
			line = line.erase(start_tag, delete_class+1);
			size_t bracket = line.rfind("}");
			if(bracket != string::npos){
				line = line.erase(bracket,1);
			}else
				throw Invalid("Missing '}' in block tag.",line);
		}else
			throw Invalid("Missing ')' in block tag.",line);
	}else
		throw Invalid("Missing block tag.",line);
	return line;
}
/*
 * Deletes brackes and parentheses.
 *
 */
string incompletes_blocks_tex(string line){
	size_t bracket = line.find("{(");
	if(bracket != std::string::npos){
		size_t parenthesis = line.find(")");
		if(parenthesis != string::npos){
			int delete_class = parenthesis - bracket;
			line = line.erase(bracket, delete_class+1);
			size_t end_bracket = line.find("}");
			if(end_bracket !=string::npos)
				throw Invalid("Unexpected '}' in incomplete block tag.",line);
		}else
			throw Invalid("Missing ')' in block tag.",line);
	}else
		throw Invalid("Missing block tag.",line);
	return line;
}
/*
 * Detects size of a table and save it into a list
 *
 */ 
list<int> info_tex(const string& file){
	list<int> list_table;
	stringstream reader(file);
	string line;
	int check = 0;
	while(getline(reader,line)){
		size_t np = line.find('<');
		if(np == string::npos){
			size_t pipe=line.find('|');
			if(pipe != string::npos){
				if(check == 0){
					int data = 0;
					int size = line.size()-1;
					while(pipe != string::npos && (line[size]=='|' or line[size-1]=='|')){
						++pipe;
						line = line.substr(pipe,line.size()-1);
						pipe = line.find("|");
						size = line.size()-1;
						++data;
					}
					--data;
					list_table.push_back(data);
					check = 1;
				}
			}else
				check = 0;
		}
	}
	return list_table;
}
